;; This should eventually be a hash table, but hash tables are not
;; documented well enough yet.

;; Key: string naming option "name".  Refers to --"name"

;; Value: #(pref-name pref-value-name type
;;          description
;;          arg-convert-func arg-verify-func
;;          default-value
;;          value-valid?
;;          value)
(define gnucash:*prefs* '())

(define (gnucash:describe-pref name value-name type description
                               default-value
                               arg-convert-func arg-verify-func)
  (let ((pref-data
         (vector name value-name type
                 description
                 arg-convert-func
                 arg-verify-func
                 default-value
                 #f
                 #f)))
    (set! gnucash:*prefs* (assoc-set! gnucash:*prefs* name pref-data))))

(define (gnucash:prefs-handle-boolean-arg pref-data args)
  ;; --arg         means #t
  ;; --arg true    means #t
  ;; --arg false   means #f
  ;; arg-verify func is ignored.

  (let ((arg (car args)))
    (if (string=? arg "false")
        (begin
          (gnucash:pref-value-set! pref-data #f)
          (cdr args))
        (begin
          (gnucash:pref-value-set! pref-data #t)
          (if (string=? arg "true")
              (cdr args)
              args)))))

(define (gnucash:prefs-handle-integer-arg pref-data args)
  ;; Makes sure the integer is valid, then calls the verify func.
  (let ((verify-func (gnucash:pref-arg-verify-func-get pref-data))
        (arg (car args)))
    
    (let ((value (string->number arg)))
      (if (not value)
          #f
          (if (not (exact? value))
              #f
              (if verify-func
                  (if (not (verify-func value))
                      #f
                      (begin
                        (gnucash:pref-value-set! pref-data value)
                        (cdr args)))
                  (begin
                    (gnucash:pref-value-set! pref-data value)
                    (cdr args))))))))


(define (gnucash:prefs-handle-string-arg pref-data args)
  (let ((verify-func (gnucash:pref-arg-verify-func-get pref-data))
        (arg (car args)))
    (if verify-func
        (if (verify-func arg)
            (begin
              (gnucash:pref-value-set! pref-data arg)
              (cdr args))              
            #f)
        (begin
          (gnucash:pref-value-set! pref-data arg)
          (cdr args)))))

(define (gnucash:prefs-show-usage)
  (display "usage: gnucash") (newline)
  (for-each
   (lambda (item)
     (display 
      (string-append
       "  --" (gnucash:pref-name-get (cdr item)) " " 
              (gnucash:pref-value-name-get (cdr item))))
     (newline)
     (display 
      (string-append
       "    description: " (gnucash:pref-description-get (cdr item))))
     (newline)
     (display
      (string-append
       "    type: " (symbol->string (gnucash:pref-type-get (cdr item)))))
     (newline)     
     (display
      "    default: ")
     (display (gnucash:pref-default-get (cdr item)))
     (newline)     
     (newline)     
     )
   
   gnucash:*prefs*))

(define (gnucash:pref-name-get pref-data) (vector-ref pref-data 0))
(define (gnucash:pref-value-name-get pref-data) (vector-ref pref-data 1))
(define (gnucash:pref-type-get pref-data) (vector-ref pref-data 2))
(define (gnucash:pref-description-get pref-data) (vector-ref pref-data 3))

(define (gnucash:pref-arg-convert-func-get pref-data) (vector-ref pref-data 4))
(define (gnucash:pref-arg-verify-func-get pref-data) (vector-ref pref-data 5))
(define (gnucash:pref-default-get pref-data) (vector-ref pref-data 6))
(define (gnucash:pref-value-valid? pref-data) (vector-ref pref-data 7))
(define (gnucash:pref-value-get pref-data)
  (if (gnucash:pref-value-valid? pref-data)
      (vector-ref pref-data 8)
      (gnucash:pref-default-get pref-data)))

(define (gnucash:pref-value-set! pref-data value)
  (vector-set! pref-data 7 #t)
  (vector-set! pref-data 8 value))

(define (gnucash:pref-lookup-def pref)
  (assoc-ref gnucash:*prefs* pref))

(define (gnucash:load-startup-config-files)
  (gnucash:debug "loading config files")

  (let* ((config-dir (gnucash:pref-value-get
                      (gnucash:pref-lookup-def "config-dir")))
         (system-config (string-append config-dir "/config"))
         (user-file
          (string-append (getenv "HOME") "/.gnucash/config.user"))
         (auto-file
          (string-append (getenv "HOME") "/.gnucash/config.auto")))
    
    (if (not (false-if-exception (load system-config)))
        (begin
          (gnucash:warn "failure loading " system-config)
          #f)
        (if (access? user-file F_OK)
            (if (not (false-if-exception (load user-file)))
                (begin
                  (gnucash:warn "failure loading " user-file)
                  #f))
            (if (access? auto-file F_OK)
                (if (not (false-if-exception (load auto-file)))
                    (begin
                      (gnucash:warn "failure loading " auto-file)
                      #f)))))))

(define (gnucash:handle-special-case-args)
  (let ((special-cases '("--debug" "--config-dir"))
        (args (program-arguments)))

    (for-each
     (lambda (case)
       (let ((result (member case args)))
         (if result
             (let* ((arg-string (make-shared-substring case 2))
                    (arg-def (gnucash:pref-lookup-def arg-string)))
               ((gnucash:pref-arg-convert-func-get arg-def)
                arg-def (cdr result))))))
     special-cases)
     (gnucash:debug "handled special case args " special-cases)))


(define (gnucash:handle-command-line-args)
  (gnucash:debug "handling command line arguments")

  (let ((args (program-arguments))
        (files-to-open '())
        (result #t))
    
    (do ((rest (cdr args))     ; the cdr skips argv[0]
         (item #f))
        ((null? rest))
      (set! item (car rest))      
      (gnucash:debug "handling arg " item)
      
      (if (not (string=? "--" (make-shared-substring item 0 2)))
          (begin
            (gnucash:debug "non-option " item ", assuming file")
            (set! rest (cdr rest))
            (set! files-to-open (cons item files-to-open)))
          
          ;; Got something that looks like an option...
          (let* ((arg-string (make-shared-substring item 2))
                 (arg-def (gnucash:pref-lookup-def arg-string)))

            (if (not arg-def)
                (begin
                  (gnucash:prefs-show-usage)
                  (set! result #f)
                  (set! rest '()))
                (let ((arg-result
                       ((gnucash:pref-arg-convert-func-get arg-def)
                        arg-def (cdr rest))))
                  (if (not arg-result)
                      (begin                
                        (set result #f)
                        (set! rest '()))
                      (set! rest arg-result)))))))
       (if result (gnucash:debug "files to open: " files-to-open))
       result))

(define (gnucash:warn . items)
  (display "gnucash: [W] ")
  (for-each (lambda (i) (display i)) items)
  (newline))

(define (gnucash:debug . items)
  (if (gnucash:pref-value-get (gnucash:pref-lookup-def "debug"))
      (begin
        (display "gnucash: [D] ")
        (for-each (lambda (i) (display i)) items)
        (newline))))
